﻿#region license
// Copyright (c) 2003, 2004, 2005 Rodrigo B. de Oliveira (rbo@acm.org)
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:
// 
//     * Redistributions of source code must retain the above copyright notice,
//     this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright notice,
//     this list of conditions and the following disclaimer in the documentation
//     and/or other materials provided with the distribution.
//     * Neither the name of Rodrigo B. de Oliveira nor the names of its
//     contributors may be used to endorse or promote products derived from this
//     software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
// CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
// OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
// THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#endregion


/*
This script visits the AST structure after each step in the compile process
and converts it to XML or back to boo syntax.
If there are visible differences since the previous step, it saves the output
to a file in a folder named "compilersteps".

How to run:

/path/to/booi.exe path/to/showsteps.boo [-xml] [-notypes] path/to/your/script.boo

You can also use the "-r:assembly.dll" flag to add assembly references.

It will generate a folder in the current directory named "compilersteps",
and put copies of the script as it looks after each compiler step in that folder.
*/

import System
import System.IO
import System.Xml.Serialization from System.Xml
import Boo.Lang.Compiler from Boo.Lang.Compiler
import Boo.Lang.Compiler.IO
import Boo.Lang.Compiler.Pipelines
import Boo.Lang.Compiler.Ast
import Boo.Lang.Compiler.Ast.Visitors
import Boo.Lang.Compiler.TypeSystem
import Boo.Lang.Compiler.Steps
import System.Reflection

[Module]
class Globals:
	static format = "boo" //"boo" //or "xml" //format for output
	static foldername = "compilersteps"
	static savefolder as string
	static n = 0
	static laststep as string
	static showtypes = true
	
class BooTypePrinterVisitor(BooPrinterVisitor):
	def constructor(writer as TextWriter):
		super(writer)
		
	override def Visit(node as Node) as bool:
		tagname = ""
		if node is not null:
			entity = TypeSystemServices.GetOptionalEntity(node)
			if entity is not null:
				tagname = entity.GetType().ToString(). \
					Replace("Boo.Lang.Compiler.TypeSystem.","")
				s = "<"
				s += tagname
				/*
				//uncomment if you want to view the expressiontype too
				exp = node as Expression
				if exp is not null and exp.ExpressionType is not null:
					s += " ExpType="+exp.ExpressionType.ToString()
				*/
				s += ">"
				Write(s)
		result = super(node)
		if tagname != "":
			Write("</"+tagname+">")
		return result
	
def PrintAST([required]result as CompilerContext, [required]o as TextWriter):
	astobject = result.CompileUnit
	try:
		s = XmlSerializer( astobject.GetType() )
	except e:
		print e.Message
		return
	try:
		s.Serialize( o, astobject )
	except e:
		print "\n", e.ToString()

def AfterStep(sender, e as CompilerStepEventArgs):
	++n
	stepname = e.Step.ToString().Replace("Boo.Lang.Parser.","").Replace("Boo.Lang.Compiler.Steps.","")
	
	tempfile = Path.GetTempFileName()
	using temp = StreamWriter(tempfile):
		if format == "xml":
			PrintAST(e.Context, temp)
		else:
			printer as BooPrinterVisitor
			if showtypes:
				printer = BooTypePrinterVisitor(temp)
			else:
				printer = BooPrinterVisitor(temp)
			printer.Print(e.Context.CompileUnit)
	
	using r = StreamReader(tempfile):
		thisstep = r.ReadToEnd()
		
	filename = string.Format("STEP{0:D2}-{1}.{2}", n, stepname, format)
	
	if thisstep != laststep:
		File.Move(tempfile, Path.Combine(savefolder, filename))
		laststep = thisstep
		print string.Format("STEP{0:D2}-{1}: SAVED TO {2} FILE.", n, stepname, format.ToUpper())
	else:
		File.Delete(tempfile)
		print string.Format("STEP{0:D2}-{1}: NO CHANGE TO AST.", n, stepname)
	
def LoadAssembly(assemblyName as string) as Assembly:
	reference as Assembly = Assembly.LoadWithPartialName(assemblyName)
	if reference is null:
		reference = Assembly.LoadFrom(Path.GetFullPath(assemblyName))
		if reference is null:
			raise ApplicationException(ResourceManager.Format("BooC.UnableToLoadAssembly", assemblyName))
	return reference

///////////////////////////////////////////////////

if len(argv) == 0:
	print "Please specify at least one boo file as a parameter"
	return

compiler = BooCompiler()

//delete old folder if running more than once:
if Directory.Exists(foldername):
	Directory.Delete(foldername, true)
	
savedir = Directory.CreateDirectory(foldername)
if savedir is null or not Directory.Exists(foldername):
	print "The directory '${foldername}' could not be created."
	return

savefolder = savedir.FullName

compiler.Parameters.Pipeline = Compile()
compiler.Parameters.Pipeline.AfterStep += AfterStep

for arg in argv:
	if arg[0:3] == "-r:":
		compiler.Parameters.References.Add(LoadAssembly(arg[3:]))
	elif arg == "-xml":
		format = "xml"
	elif arg == "-notypes":
		showtypes = false
	else:
		compiler.Parameters.Input.Add(FileInput(arg))

try:
	print "\nSee src/Boo.Lang.Compiler/Steps/ for the source code for these steps:\n"
	result = compiler.Run()
	if len(result.Errors) > 0:
		print "\nThere were errors compiling the boo file(s)"
	else:
		print "\nSuccessful: See the files under: '${savefolder}'"
except e:
	print e.GetType(), ":", e.Message
